package main

import (
	"bufio"
	"crypto"
	"crypto/ed25519"
	"crypto/rand"
	"encoding/base64"
	"errors"
	"fmt"
	"os"
	"time"

	"github.com/go-piv/piv-go/v2/piv"

	ed25519_tor "github.com/cretz/bine/torutil/ed25519"
)

const DATEFMT = "2006-01-02"

const IDENTITY_PK_HEADER = "== ed25519v1-public: type0 ==\x00\x00\x00"
const SIGNING_SK_HEADER = "== ed25519v1-secret: type4 ==\x00\x00\x00"
const SIGNING_CERT_HEADER = "== ed25519v1-cert: type4 ==\x00\x00\x00\x00\x00"

// Fetch the identity public key from the Yubikey and verify it is Ed25519
func FetchIdentityPk(yk *piv.YubiKey, slot piv.Slot) (ed25519.PublicKey, error) {
	cert, err := yk.Attest(slot)
	if err != nil {
		return nil, err
	}
	pk, ok := cert.PublicKey.(ed25519.PublicKey)
	if !ok {
		return nil, errors.New("public key is not ed25519.PublicKey")
	}
	return pk, nil
}

// Fetch the identity secret key from the Yubikey as a signer
func FetchIdentitySk(yk *piv.YubiKey, slot piv.Slot, pk ed25519.PublicKey) (crypto.Signer, error) {
	keyAuth := piv.KeyAuth{
		PINPrompt: PromptPin,
		PINPolicy: piv.PINPolicyAlways,
	}

	sk, err := yk.PrivateKey(slot, pk, keyAuth)
	if err != nil {
		return nil, err
	}

	signer, ok := sk.(crypto.Signer)
	if !ok {
		return nil, errors.New("crypto.Signer is not implemented")
	}

	return signer, nil
}

func main() {
	scanner := bufio.NewScanner(os.Stdin)

	// Obtain the list of all connected smart cards
	cards, err := piv.Cards()
	if err != nil {
		panic(err)
	}
	if len(cards) == 0 {
		panic("no smart cards found")
	}

	// Prompt for the smart card
	var card string
	for {
		card, err = PromptCard(scanner, cards)
		if err == nil {
			break
		} else {
			fmt.Fprintln(os.Stderr, err)
		}
	}

	// Open the smart card
	yk, err := piv.Open(card)
	if err != nil {
		panic(err)
	}
	defer yk.Close()

	var slot piv.Slot
	for {
		slot, err = PromptSlot(scanner, yk)
		if err == nil {
			break
		} else {
			fmt.Fprintln(os.Stderr, err)
		}
	}

	// Fetch the keys
	identityPk, err := FetchIdentityPk(yk, slot)
	if err != nil {
		panic(err)
	}
	identitySk, err := FetchIdentitySk(yk, slot, identityPk)
	if err != nil {
		panic(err)
	}

	// Fetch the serial number
	serial, err := yk.Serial()
	if err != nil {
		panic(err)
	}

	// Yubikey setup is done now, accept parameters about the certificate

	// Prompt for the expiration date
	var expirationDate time.Time
	for {
		expirationDate, err = PromptExpirationDate(scanner)
		if err == nil {
			break
		} else {
			fmt.Fprintln(os.Stderr, err)
		}
	}

	// Prompt for the paths
	identityPkPath := PromptPath(scanner, "identity public key", DEFAULT_IDENTITY_PK_PATH)
	signingSkPath := PromptPath(scanner, "signing secret key", DEFAULT_SIGNING_SK_PATH)
	signingCertPath := PromptPath(scanner, "signing cert", DEFAULT_SIGNING_CERT_PATH)

	// Now verify the user input
	fmt.Println()
	fmt.Println("================================")
	fmt.Printf("smart card:\t\t\t%s %d\n", card, serial)
	fmt.Printf("slot:\t\t\t\t0x%x\n", slot.Key)
	fmt.Printf("public key:\t\t\t%s\n", base64.RawStdEncoding.EncodeToString(identityPk))
	fmt.Printf("expiration date:\t\t%s (%d)\n", expirationDate.Format(time.RFC1123), expirationDate.Unix())
	fmt.Printf("identity public key path:\t%s\n", identityPkPath)
	fmt.Printf("signing secret key path:\t%s\n", signingSkPath)
	fmt.Printf("signing cert path:\t\t%s\n", signingCertPath)
	fmt.Println("================================")
	fmt.Println()

	// Ask for a final confirmation
	if !PromptContinue(scanner) {
		fmt.Fprintln(os.Stderr, "user aborted")
		os.Exit(1)
	}

	// Generate the signing key
	signingKp, err := ed25519_tor.GenerateKey(rand.Reader)
	if err != nil {
		panic(err)
	}

	// Create and sign the certificate
	cert := TorCert{ExpirationDate: expirationDate, SigningKey: signingKp.PublicKey(), IdentityKey: ed25519_tor.FromCryptoPublicKey(identityPk)}
	sig, err := (identitySk).Sign(rand.Reader, cert.EncodeNoSignature(), crypto.Hash(0))
	if err != nil {
		panic(err)
	}
	copy(cert.Signature[:], sig[:64])

	// Export all files
	err = CtorExport(identityPkPath, IDENTITY_PK_HEADER, ed25519_tor.FromCryptoPublicKey(identityPk))
	if err != nil {
		panic(err)
	}
	err = CtorExport(signingSkPath, SIGNING_SK_HEADER, signingKp.PrivateKey())
	if err != nil {
		panic(err)
	}
	err = CtorExport(signingCertPath, SIGNING_CERT_HEADER, cert.Encode())
	if err != nil {
		panic(err)
	}

	// We are done :)
	fmt.Println("success! 🧅")
}
