package main

import (
	"bufio"
	"errors"
	"fmt"
	"os"
	"strconv"
	"time"

	"github.com/go-piv/piv-go/v2/piv"
	"golang.org/x/term"
)

const DEFAULT_IDENTITY_PK_PATH = "ed25519_master_id_public_key"
const DEFAULT_SIGNING_SK_PATH = "ed25519_signing_secret_key"
const DEFAULT_SIGNING_CERT_PATH = "ed25519_signing_cert"

// Prompt the user to choose a smart card
func PromptCard(scanner *bufio.Scanner, cards []string) (string, error) {
	// Print a list of all smart cards
	for i, name := range cards {
		fmt.Printf("%d. %s\n", i+1, name)
	}

	fmt.Print("choose a smart card: ")
	scanner.Scan()
	i, err := strconv.Atoi(scanner.Text())
	if err != nil {
		return "", errors.New("input is not a number")
	}
	i -= 1

	if i < 0 || i >= len(cards) {
		return "", errors.New("smart card not found")
	}

	return cards[i], nil
}

// Prompt the user to choose a slot, defaults to `piv.SlotSignature`
func PromptSlot(scanner *bufio.Scanner, yk *piv.YubiKey) (piv.Slot, error) {
	slot := piv.SlotSignature

	fmt.Print("choose a slot [9c]: ")
	scanner.Scan()
	s := scanner.Text()
	if s == "" {
		s = "9c"
	}

	i, err := strconv.ParseUint(s, 16, 8)
	if err != nil {
		return slot, err
	}

	switch i {
	case 0x9a:
		slot = piv.SlotAuthentication
	case 0x9c:
		slot = piv.SlotSignature
	case 0x9e:
		slot = piv.SlotCardAuthentication
	case 0x9d:
		slot = piv.SlotKeyManagement
	default:
		var ok bool
		slot, ok = piv.RetiredKeyManagementSlot(uint32(i))
		if !ok {
			return slot, errors.New("slot does not exist")
		}
	}

	// Call KeyInfo to check whether the key actually exists on the slot
	_, err = yk.KeyInfo(slot)

	return slot, err
}

// Prompt the user to choose an expiration while providing a default value
func PromptExpirationDate(scanner *bufio.Scanner) (time.Time, error) {
	expiration_date := time.Date(time.Now().Year(), time.Now().Month(), time.Now().Day(), 0, 0, 0, 0, time.UTC).AddDate(0, 3, 0)
	fmt.Printf("choose an expiration date [%s]: ", expiration_date.Format(DATEFMT))
	scanner.Scan()
	s := scanner.Text()

	// Use the default
	if s == "" {
		return expiration_date, nil
	}

	res, err := time.Parse(DATEFMT, s)
	return res, err
}

// Prompt the user to choose a path while providing a default value
func PromptPath(scanner *bufio.Scanner, name string, def string) string {
	fmt.Printf("choose a path for the %s [%s]: ", name, def)
	scanner.Scan()
	s := scanner.Text()

	if s == "" {
		return def
	} else {
		return s
	}
}

// Prompt the user for a (y/n)
func PromptContinue(scanner *bufio.Scanner) bool {
	fmt.Print("is this correct? (y/n): ")
	scanner.Scan()
	s := scanner.Text()

	if s == "y" || s == "Y" {
		return true
	} else {
		return false
	}
}

// A simple password read from stdin
func PromptPin() (string, error) {
	fmt.Print("enter PIN: ")
	pin, err := term.ReadPassword(int(os.Stdin.Fd()))
	if err != nil {
		return "", err
	}
	fmt.Println()
	return string(pin), nil
}
